l :- consult('vgrad.pl').

% Hilfsprdikate
% app(L1,L2,L3): Vereinige L1 und L2 zu L3 unter Ausschluss von Duplikaten
app([],L,L).
app([X|L],L1,L2) :- member(X,L1), !, app(L,L1,L2).
app([X|L],L1,[X|L2]) :- app(L,L1,L2).

% member(X,L): Wenn X in der Liste L enthalten ist
member(X,[X|_]).
member(X,[_|RL]) :- member(X,RL).

% subtract(L1,L2,L3): L3 bildet die Differenzmenge von L1 und L2
subtract([],_,[]).
subtract([X|L],L1,L2) :- member(X,L1), !, subtract(L,L1,L2).
subtract([X|L],L1,[X|L2]) :- subtract(L,L1,L2).

% append(L1,L2,L3): L3 bildet die Verkettung von L1 und L2
append([],L,L).
append([X|L1],L2,[X|L3]) :- append(L1,L2,L3).

% Ehe-Prdikat
% ehe(X,Y): X ist mit Y verheiratet
ehe(uwe,helga).
ehe(henry,heike).
ehe(erich,marion).
ehe(heinrich,rosa).
ehe(andreas,katrin).
ehe(dora,max).
ehe(erna,ernst).
ehe(ilona,norbert).
ehe(helmut,else).
ehe(hans,steffi).
ehe(erhard,mia).
ehe(jana,carsten).

% Kind-Prdikat
% kind(X,Y): X ist Kind von Y
kind(heike,uwe).
kind(andreas,uwe).
kind(dora,henry).
kind(max,erich).
kind(ernst,andreas).
kind(erna,heinrich).
kind(elvira,max).
kind(ilona,max).
kind(norbert,ernst).
kind(ursula,ernst).
kind(dieter,ernst).
kind(helmut,ernst).
kind(matthias,norbert).
kind(michael,norbert).
kind(jana,norbert).
kind(steffi,helmut).
kind(erhard,helmut).
kind(anna,hans).
kind(lena,hans).
kind(carsten,erhard).
kind(maria,carsten).

% Ehe ist kommutativ
ehe2(X,Y) :- ehe(X,Y).
ehe2(X,Y) :- ehe(Y,X).

% X ist auch Kind vom Ehepartner
kind2(X,Y) :- kind(X,Y).
kind2(X,Y) :- ehe2(Y,Z), kind(X,Z).


% Verwandschaften 1. Grades
dverw(X,Y) :- ehe2(X,Y).                        % X und Y verheiratet
dverw(X,Y) :- kind2(X,Y).                       % X ist Kind von Y
dverw(X,Y) :- kind2(Y,X).                       % X ist Elternteil von Y
dverw(X,Y) :- kind2(X,Z), !, kind2(Y,Z), X\==Y. % X und Y sind Geschwister, Cut gegen Duplikate

% breitensuche([X],Y,G,_): Berechne den Verwandtschaftsgrad G von X zu Y
% Rekursionsaufrufe von breitensuche(L1,Y,G,L2): Berechne die Liste der im 1. Grad Verwandten aller Elemente aus L1 unter Ausschluss der Elemente aus L2
breitensuche(VerwL,Ziel,0,_) :- member(Ziel,VerwL).   % Abbruch, wenn die Zielperson Element der Liste der nchsten direkt Verwandten ist
breitensuche(VerwL,Ziel,Grad,BetrL) :-
    VerwL \== [],                                     % Liste der zu betrachtenden Personen darf nicht leer sein
    gradNaechst(VerwL,Verw1L),                        % erhalte Liste Verw1L der mit den Personen in VerwL im 1. Gad Verwandten
    subtract(Verw1L,BetrL,Verw2L),                    % schliee in BetrL enthaltene Elemente aus Verw1L aus
    append(BetrL,Verw2L,BetrNeuL),                    % Elemente aus Verw2L sollen nicht noch einmal betrachtet werden
    breitensuche(Verw2L,Ziel,Grad1,BetrNeuL),         % berechne Liste der im 1. Grad Verwandten von Verw2L
    Grad is Grad1+1.                                  % 1 Rekursionsaufruf == 1 Verwandtschaftsgrad
    
% gradNaechst(L1,L2): L2 gibt die Liste der Personen an, die mit denen aus L1 im 1. Grad verwandt sind
gradNaechst([],[]).                                   % Abbruchbedingung
gradNaechst(VerwL,NaechstVerwL) :-
    VerwL=[Erster|Restliste],                         % Erster ist erste Person der bergebenen Liste
    setof(Verwandte,                                  % erhalte die Verwandten ...
        dverw(Erster,Verwandte),                      % ... die direkt mit 'Erster' verwandt sind ...
        Verw1L),                                      % ... und schreibe das Ergebnis in Verw1L
    gradNaechst(Restliste,Verw2L),                    % rufe rekursiv fr Restliste auf
    app(Verw1L,Verw2L,NaechstVerwL).                  % verkette die Ergebnislisten unter Ausschluss von Duplikaten

% verwandt(X,Y,G): X ist mit Y im Grad G verwandt
% rufe breitensuche() auf, berechne Verwandtschaftsgrad von Start zu Ziel; Cut verhindert Rekursion
verwandt(Start,Ziel,Grad) :- breitensuche([Start],Ziel,Grad,[Start]).
